/*
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*global define, console, navigator, window, setTimeout, tizen*/
/*jslint regexp: true*/

/**
 * Camera model module.
 *
 * @requires core/event
 * @requires core/storage/localstorage
 * @requires helpers/date
 * @namespace models/camera
 * @memberof models
 */

define({
    name: 'models/camera',
    requires: [
        'core/event',
        'core/storage/localstorage',
        'helpers/date'
    ],
    def: function modelsCamera(req) {
        /*jshint maxstatements:48*/
        'use strict';

        /**
         * Core event module.
         *
         * @private
         * @type {Module}
         */
        var e = req.core.event,

            /**
             * Core local storage module.
             *
             * @private
             * @type {Module}
             */
            storage = req.core.storage.localstorage,

            /**
             * Date helper module.
             *
             * @private
             * @type {Module}
             */
            dateHelper = req.helpers.date,

            /**
             * Max recording time.
             *
             * @public
             * @const {number}
             */
            MAX_RECORDING_TIME = 10000,

            /**
             * Interval time for checking video length.
             *
             * @private
             * @const {number}
             */
            VIDEO_LENGTH_CHECK_INTERVAL = 10,

            /**
             * Autofocus delay time.
             *
             * @private
             * @const {number}
             */
            AUTOFOCUS_DELAY = 500,

            /**
             * Taking picture delay.
             *
             * @private
             * @const {number}
             */
            TAKE_PICTURE_DELAY = 100,

            /**
             * Path to pictures destination directory.
             *
             * @private
             * @type {string}
             */
            pictureDestinationDirectory = '',

            /**
             * Path to videos destination directory.
             *
             * @private
             * @type {string}
             */
            videoDestinationDirectory = '',

            /**
             * Storage setting key.
             *
             * @private
             * @const {string}
             */
            STORAGE_SETTINGS_KEY = 'settings',

            /**
             * Camera control object.
             *
             * @private
             * @type {CameraControl}
             */
            cameraControl = null,

            /**
             * Path to the new picture.
             *
             * @private
             * @type {string}
             */
            picturePath = '',

            /**
             * Path to the new recording.
             *
             * @private
             * @type {string}
             */
            videoPath = '',

            /**
             * Camera settings object.
             *
             * @private
             * @type {object}
             */
            cameraSettings = {},

            /**
             * Unique interval ID for video length check.
             *
             * @private
             * @type {number}
             */
            videoLengthCheckInterval = null,

            /**
             * Time of recording start.
             *
             * @private
             * @type {Date}
             */
            videoRecordingStartTime = null,

            /**
             * Length of recording current length (in milliseconds).
             *
             * @private
             * @type {number}
             */
            videoRecordingTime = 0,

            /**
             * Flag indicating if camera is recording or taking picture.
             *
             * @private
             * @type {boolean}
             */
            busy = false,

            /**
             * Flag indicating if camera is taking picture.
             *
             * @private
             * @type {boolean}
             */
            takingPictureInProgress = false,

            /**
             * Flag indicating if taking picture cancellation request was made.
             *
             * @private
             * @type {boolean}
             */
            takingPictureCancelRequested = false;

        /**
         * Handles camera onshutter event.
         *
         * @private
         * @fires models.camera.shutter
         */
        function onShutter() {
            e.fire('shutter');
        }

        /**
         * Returns true if camera is ready to work, false otherwise.
         *
         * @memberof models/camera
         * @public
         * @returns {boolean}
         */
        function isReady() {
            return cameraControl !== null;
        }

        /**
         * Returns true if camera is recording, false otherwise.
         *
         * @memberof models/camera
         * @public
         * @returns {boolean}
         */
        function isRecording() {
            return !!videoLengthCheckInterval;
        }

        /**
         * Saves camera settings to storage.
         *
         * @private
         */
        function saveCameraSettings() {
            storage.add(STORAGE_SETTINGS_KEY, cameraSettings);
        }

        /**
         * Returns array of available picture sizes.
         *
         * @memberof models/camera
         * @public
         * @returns {CameraSize[]}
         */
        function getAvailablePictureSizes() {
            return cameraControl.capabilities.pictureSizes;
        }

        /**
         * Returns array of available picture formats.
         *
         * @memberof models/camera
         * @public
         * @returns {string[]}
         */
        function getAvailablePictureFormats() {
            return cameraControl.capabilities.pictureFormats;
        }

        /**
         * Returns array of available recording formats.
         *
         * @memberof models/camera
         * @public
         * @returns {string[]}
         */
        function getAvailableRecordingFormats() {
            return cameraControl.capabilities.recordingFormats;
        }

        /**
         * Returns picture format.
         *
         * @memberof models/camera
         * @public
         * @returns {string}
         */
        function getPictureFormat() {
            return cameraSettings.pictureFormat;
        }

        /**
         * Returns recording format.
         *
         * @memberof models/camera
         * @public
         * @returns {array}
         */
        function getRecordingFormat() {
            return cameraSettings.recordingFormat;
        }

        /**
         * Returns current recording time in milliseconds.
         *
         * @memberof models/camera
         * @public
         * @returns {number}
         */
        function getRecordingTime() {
            return videoRecordingTime;
        }

        /**
         * Returns current picture size.
         *
         * @memberof models/camera
         * @public
         * @returns {object}
         */
        function getPictureSize() {
            return {
                width: cameraSettings.pictureSize.width,
                height: cameraSettings.pictureSize.height
            };
        }

        /**
         * Sets picture format.
         *
         * @memberof models/camera
         * @public
         * @param {string} format
         */
        function setPictureFormat(format) {
            cameraSettings.pictureFormat = format;
            saveCameraSettings();
        }

        /**
         * Sets recording format.
         *
         * @memberof models/camera
         * @public
         * @param {string} format
         */
        function setRecordingFormat(format) {
            cameraSettings.recordingFormat = format;
            saveCameraSettings();
        }

        /**
         * Sets picture size.
         *
         * @memberof models/camera
         * @public
         * @param {object} size
         */
        function setPictureSize(size) {
            cameraSettings.pictureSize = {
                width: size.width,
                height: size.height
            };
            saveCameraSettings();
        }

        /**
         * Executes on camera control creation error.
         *
         * @private
         * @param {object} error
         * @fires models.camera.error
         */
        function onCameraControlError(error) {
            console.error(error);
            e.fire('error', {error: error});
        }

        /**
         * Initializes camera settings.
         *
         * @private
         */
        function initCameraSettings() {
            var pictureFormats = null,
                pictureSizes = null,
                recordingFormats = null;

            cameraSettings = storage.get(STORAGE_SETTINGS_KEY);

            if (!cameraSettings) {
                cameraSettings = {};

                pictureFormats = getAvailablePictureFormats();
                cameraSettings.pictureFormat = pictureFormats[0];

                pictureSizes = getAvailablePictureSizes();
                cameraSettings.pictureSize = pictureSizes[0];

                recordingFormats = getAvailableRecordingFormats();
                cameraSettings.recordingFormat = recordingFormats[0];

                saveCameraSettings();
            }
        }

        /**
         * Executes when camera control is created from stream.
         *
         * @private
         * @param {CameraControl} control
         * @fires models.camera.ready
         */
        function onCameraControlCreated(control) {
            cameraControl = control;
            initCameraSettings();
            e.fire('ready');
        }

        /**
         * Registers stream that camera controls.
         *
         * @memberof models/camera
         * @public
         * @param {LocalMediaStream} mediaStream
         */
        function registerStream(mediaStream) {
            navigator.tizCamera.createCameraControl(
                mediaStream,
                onCameraControlCreated,
                onCameraControlError
            );
        }

        /**
         * Returns true if taking picture is in progress, false otherwise.
         *
         * @private
         * @returns {boolean}
         */
        function isTakingPicture() {
            return takingPictureInProgress;
        }

        /**
         * Cancels taking picture process.
         *
         * @private
         * @fires models.camera.picture.canceled
         */
        function cancelTakingPicture() {
            if (isTakingPicture()) {
                takingPictureInProgress = false;
                takingPictureCancelRequested = true;
                busy = false;
                e.fire('picture.canceled');
            }
        }

        /**
         * Executes when picture is ready.
         * Sets camera busy flag and takingPictureProgress flag to false.
         * Fires 'models.camera.picture.done' event.
         *
         * @private
         * @fires models.camera.picture.done
         */
        function onPictureDone() {
            busy = false;
            takingPictureInProgress = false;
            e.fire('picture.done', {path: picturePath});
        }

        /**
         * Executes when error occurs during taking picture.
         * Sets camera busy flag and takingPictureProgress flag to false.
         * Sets taking picture in progress helper property to false.
         * Fires 'models.camera.picture.error' event.
         *
         * @private
         * @param {object} error
         * @fires models.camera.picture.error
         */
        function onPictureError(error) {
            busy = false;
            takingPictureInProgress = false;
            e.fire('picture.error', {error: error});
        }

        /**
         * Executes when image settings are applied.
         * Checks if taking picture was canceled and if not
         * cameraControl.image.takePicture method is invoked to capture
         * an image.
         *
         * @private
         */
        function onImageSettingsApplied() {
            if (!takingPictureCancelRequested) {
                cameraControl.image.takePicture(onPictureDone, onPictureError);
            }
        }

        /**
         * Executes when error occurs during applying image settings.
         * Sets camera busy flag and takingPictureProgress flag to false.
         * Fires 'models.camera.picture.error' event.
         *
         * @private
         * @param {object} error
         * @fires models.camera.picture.error
         */
        function onImageSettingsError(error) {
            busy = false;
            takingPictureInProgress = false;
            e.fire('picture.error', {error: error});
        }

        /**
         * Creates filename for new picture.
         *
         * @private
         * @returns {string}
         */
        function createPictureFileName() {
            var currentDate = new Date(),
                pictureFormat = getPictureFormat(),
                extension = pictureFormat === 'jpeg' ? 'jpg' : pictureFormat,
                fileName = '';

            fileName = dateHelper.format(currentDate, 'yyyymmdd_HHMMSS') +
                '.' + extension;

            return fileName;
        }

        /**
         * Starts taking photo.
         *
         * @private
         */
        function startTakingPicture() {
            var settings = {},
                fileName = '';

            fileName = createPictureFileName();
            picturePath = pictureDestinationDirectory + '/' + fileName;

            settings.fileName = fileName;
            settings.pictureFormat = getPictureFormat();
            settings.pictureSize = getPictureSize();

            if (!cameraControl) {
                cancelTakingPicture();
                return;
            }
            cameraControl.image.onshutter = onShutter;
            cameraControl.image.applySettings(
                settings,
                onImageSettingsApplied,
                onImageSettingsError
            );
        }

        /**
         * Handles autofocus success.
         * Fires 'models.camera.autofocus.success' event.
         * Checks if taking picture was canceled and if not startTakingPicture
         * method is invoked with specified delay.
         *
         * @private
         * @fires models.camera.autofocus.success
         */
        function onAutoFocusSuccess() {
            e.fire('autofocus.success');
            if (!takingPictureCancelRequested) {
                setTimeout(startTakingPicture, TAKE_PICTURE_DELAY);
            }

        }

        /**
         * Handles autofocus error.
         * Fires 'models.camera.autofocus.failure' event.
         * Checks if taking picture was canceled and if not retries
         * startTakingPicture method to invoke with specified delay.
         *
         * @private
         * @fires models.camera.autofocus.failure
         */
        function onAutoFocusFailure() {
            e.fire('autofocus.failure');
            if (!takingPictureCancelRequested) {
                setTimeout(startTakingPicture, TAKE_PICTURE_DELAY);
            }
        }

        /**
         * Takes a picture.
         * When autofocus starts, camera.autofocus event is fired.
         * When autofocus succeeds, camera.autofocus.success event is fired.
         * When autofocus fails, camera.autofocus.failure event is fired.
         * When picture is ready camera.picture.done event is fired with picture
         * path as a data. If error occurs, camera.picture.error event is fired.
         *
         * @memberof models/camera
         * @public
         * @returns {boolean} If process starts true is returned, false
         * otherwise (camera other operation is in progress).
         * @fires models.camera.autofocus.start
         */
        function takePicture() {
            if (busy) {
                return false;
            }

            busy = true;
            takingPictureInProgress = true;
            takingPictureCancelRequested = false;
            e.fire('autofocus.start');
            if (cameraControl.autoFocus()) {
                setTimeout(onAutoFocusSuccess, AUTOFOCUS_DELAY);
            } else {
                setTimeout(onAutoFocusFailure, AUTOFOCUS_DELAY);
            }

            return true;
        }

        /**
         * Stops tracing video length.
         *
         * @private
         */
        function stopTracingVideoLength() {
            window.clearInterval(videoLengthCheckInterval);
            videoLengthCheckInterval = null;
        }

        /**
         * Executes when error occurs during applying video settings.
         *
         * @private
         * @param {object} error
         * @fires models.camera.recording.error
         */
        function onVideoSettingsError(error) {
            console.error('settings.error');
            busy = false;
            e.fire('recording.error', {error: error});
        }

        /**
         * Creates filename for new video.
         *
         * @private
         * @returns {string}
         */
        function createVideoFileName() {
            var currentDate = new Date(),
                extension = getRecordingFormat(),
                fileName = '';

            fileName = dateHelper.format(currentDate, 'yyyymmdd_HHMMSS') +
                '.' + extension;

            return fileName;
        }

        /**
         * Executes when video recording stops successfully.
         *
         * @private
         * @fires models.camera.recording.done
         */
        function onVideoRecordingStopSuccess() {
            busy = false;
            e.fire('recording.done', {path: videoPath});
            videoRecordingTime = 0;
        }

        /**
         * Executes when video recording stop fails.
         *
         * @private
         * @param {object} error
         * @fires models.camera.recording.error
         */
        function onVideoRecordingStopError(error) {
            busy = false;
            e.fire('recording.error', {error: error});
            videoRecordingTime = 0;
        }

        /**
         * Stop video recording.
         * When recording is stopped, camera.recording.done event is fired
         * with file path as a data.
         * If error occurs camera.recording error is fired.
         * Recording will stop also if MAX_RECORDING_TIME will be reached.
         *
         * @memberof models/camera
         * @public
         */
        function stopRecording() {
            stopTracingVideoLength();
            cameraControl.recorder.stop(onVideoRecordingStopSuccess,
                onVideoRecordingStopError
            );
        }


        /**
         * Checks if video length is greater then MAX_RECORDING_TIME.
         * If it does, recording will be stopped.
         *
         * @private
         */
        function checkVideoLength() {
            var currentTime = new Date();

            videoRecordingTime = currentTime - videoRecordingStartTime;
            if (videoRecordingTime > MAX_RECORDING_TIME) {
                stopRecording();
            }
        }

        /**
         * Starts tracing video length.
         * When video length reaches MAX_RECORDING_TIME, recording
         * will be stopped automatically.
         *
         * @private
         */
        function startTracingVideoLength() {
            videoRecordingStartTime = new Date();
            videoLengthCheckInterval = window.setInterval(
                checkVideoLength,
                VIDEO_LENGTH_CHECK_INTERVAL
            );
        }

        /**
         * Executes when recording starts successfully.
         *
         * @private
         * @fires models.camera.recording.start
         */
        function onRecordingStartSuccess() {
            startTracingVideoLength();
            e.fire('recording.start');
        }

        /**
         * Executes when error occurs during recording start.
         *
         * @private
         * @param {object} error
         * @fires models.camera.recording.error
         */
        function onRecordingStartError(error) {
            busy = false;
            e.fire('recording.error', {error: error});
        }

        /**
         * Executes when video settings are applied.
         *
         * @private
         */
        function onVideoSettingsApplied() {
            cameraControl.recorder.start(
                onRecordingStartSuccess,
                onRecordingStartError
            );
        }

        /**
         * Releases camera.
         *
         * @memberof models/camera
         * @public
         * @fires models.camera.release
         */
        function release() {
            if (isRecording()) {
                stopRecording();
            } else if (isTakingPicture()) {
                cancelTakingPicture();
            }
            busy = false;
            if (cameraControl) {
                cameraControl.release();
                cameraControl = null;
                e.fire('release');
            }
        }

        /**
         * Starts video recording.
         * When recording is started successfully, camera.recording.start event
         * is fired. If error occurs, camera.recording.error event is fired.
         *
         * @memberof models/camera
         * @public
         * @returns {boolean} If process starts true is returned, false
         * otherwise (camera other operation is in progress).
         */
        function startRecording() {
            var settings = {},
                fileName = '';

            if (busy) {
                return false;
            }

            busy = true;
            fileName = createVideoFileName();
            videoPath = videoDestinationDirectory + '/' + fileName;

            settings.fileName = fileName;
            settings.recordingFormat = getRecordingFormat();

            cameraControl.recorder.applySettings(
                settings,
                onVideoSettingsApplied,
                onVideoSettingsError
            );

            return true;
        }

        /**
         * Resolves photos destination path.
         */
        function resolvePhotosDestination() {
            try {
                tizen.filesystem.resolve(
                    'images',
                    function onImagesResolved(file) {
                        pictureDestinationDirectory = file.toURI()
                            .replace('file://', '');
                    },
                    function onImagesResolveError(error) {
                        console.error('images resolve error', error);
                    },
                    'r'
                );
            } catch (error) {
                console.error('images resolve error', error);
            }
        }

        /**
         * Resolves videos destination path.
         */
        function resolveVideosDestination() {
            try {
                tizen.filesystem.resolve(
                    'videos',
                    function onVideosResolved(file) {
                        videoDestinationDirectory = file.toURI()
                            .replace('file://', '');
                    },
                    function onVideosResolveError(error) {
                        console.error('videos resolve error', error);
                    },
                    'r'
                );
            } catch (error) {
                console.error('videos resolve error', error);
            }
        }

        /**
         * Returns true if Tizen Camera API is supported on the device,
         * false otherwise.
         *
         * @memberof models/camera
         * @public
         * @returns {boolean}
         */
        function isSupported() {
            return !!(navigator && navigator.tizCamera);
        }

        /**
         * Initializes module.
         * @public
         */
        function init() {
            resolvePhotosDestination();
            resolveVideosDestination();
        }

        return {
            MAX_RECORDING_TIME: MAX_RECORDING_TIME,

            getAvailablePictureSizes: getAvailablePictureSizes,
            getPictureSize: getPictureSize,
            setPictureSize: setPictureSize,

            getAvailablePictureFormats: getAvailablePictureFormats,
            getPictureFormat: getPictureFormat,
            setPictureFormat: setPictureFormat,

            getAvailableRecordingFormats: getAvailableRecordingFormats,
            getRecordingFormat: getRecordingFormat,
            setRecordingFormat: setRecordingFormat,

            registerStream: registerStream,
            release: release,
            isReady: isReady,
            isRecording: isRecording,

            takePicture: takePicture,

            startRecording: startRecording,
            stopRecording: stopRecording,
            getRecordingTime: getRecordingTime,

            isSupported: isSupported,
            init: init
        };
    }
});
