/*
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* global define, console, document, navigator, window, setInterval,
 clearInterval, tau, setTimeout */
/*jslint plusplus: true*/

/**
 * Main page module.
 *
 * @requires core/event
 * @requires core/application
 * @requires core/power
 * @requires models/camera
 * @requires helpers/adjust
 * @requires helpers/page
 * @requires helpers/date
 * @namespace views/main
 * @memberof views
 */

define({
    name: 'views/main',
    requires: [
        'core/event',
        'core/application',
        'core/power',
        'models/camera',
        'helpers/adjust',
        'helpers/page',
        'helpers/date'
    ],
    def: function viewsMain(req) {
        /*jshint maxstatements:68*/
        'use strict';

        /**
         * Core event module.
         *
         * @private
         * @type {Module}
         */
        var e = req.core.event,

            /**
             * Core power module.
             *
             * @private
             * @type {Module}
             */
            power = req.core.power,

            /**
             * Core application module.
             *
             * @private
             * @type {Module}
             */
            app = req.core.application,

            /**
             * Camera models module.
             *
             * @private
             * @type {Module}
             */
            camera = req.models.camera,

            /**
             * Adjust helpers module.
             *
             * @private
             * @type {Module}
             */
            adjustHelper = req.helpers.adjust,

            /**
             * Date helpers module.
             *
             * @private
             * @type {Module}
             */
            dateHelper = req.helpers.date,

            /**
             * Page helpers module.
             *
             * @private
             * @type {Module}
             */
            pageHelper = req.helpers.page,

            /**
             * Network "no source state" of the fetching of media over the
             * network.
             *
             * @private
             * @const {number}
             */
            VIDEO_NETWORK_NO_SOURCE = 3,

            /**
             * Error name for no free space for write a file on the device.
             *
             * @private
             * @const {string}
             */
            ERROR_FILE_WRITE = 'FILE_WRITE_ERR',

            /**
             * Message, which is shown when access to camera stream in not
             * possible.
             *
             * @private
             * @const {string}
             */
            CANNOT_ACCESS_CAMERA_MSG = 'Cannot access camera stream. ' +
                'Please close all applications that use the camera and ' +
                'open the application again.',

            /**
             * Message, which is shown when camera in not available.
             *
             * @private
             * @const {string}
             */
            CAMERA_NOT_AVAILABLE_MSG = 'Camera is not available.',

            /**
             * Message, which is shown when application is launched on the
             * emulator.
             *
             * @private
             * @const {string}
             */
            CAMERA_ON_DEVICE_ONLY_MSG = 'Camera is supported ' +
                'on device target only.',

            /**
             * Message, which is shown when there is no free space on the
             * device.
             *
             * @private
             * @const {string}
             */
            NO_FREE_SPACE_MSG = 'No free space.',

            /**
             * Message, which is shown when other not definite error occurred.
             *
             * @private
             * @const {string}
             */
            DEFAULT_ERROR_MSG = 'An error occurred.',

            /**
             * Number of attempts to obtain to access to the camera stream.
             *
             * @private
             * @const {number}
             */
            PREVIEW_INIT_MAX_ATTEMPTS = 3,

            /**
             * Time between attempts to get access to the camera stream.
             *
             * @private
             * @const {number}
             */
            PREVIEW_INIT_ATTEMPT_TIMEOUT = 500,

            /**
             * Video state when no information is available about the media
             * resource.
             *
             * @private
             * @const {number}
             */
            READY_STATE_HAVE_NOTHING = 0,

            /**
             * Interval time for updating recording progress bar.
             *
             * @private
             * @const {number}
             */
            RECORDING_INTERVAL_STEP = 100,

            /**
             * Reference to the application main page.
             *
             * @private
             * @type {HTMLElement}
             */
            page = null,

            /**
             * Reference to the change mode button.
             *
             * @private
             * @type {HTMLElement}
             */
            modeBtn = null,

            /**
             * Reference to the settings button.
             *
             * @private
             * @type {HTMLElement}
             */
            settingsBtn = null,

            /**
             * Reference to the HTML video element.
             *
             * @private
             * @type {HTMLElement}
             */
            cameraPreview = null,

            /**
             * Reference to the HTML container for visual focus frame.
             *
             * @private
             * @type {HTMLElement}
             */
            focusContainer = null,

            /**
             * Reference to the focus frame HTML element.
             *
             * @private
             * @type {HTMLElement}
             */
            focusFrame = null,

            /**
             * Reference to the HTML container for progress bar.
             *
             * @private
             * @type {HTMLElement}
             */
            recordProgress = null,

            /**
             * Reference to the HTML element for progress bar value.
             *
             * @private
             * @type {HTMLElement}
             */
            recordProgressVal = null,

            /**
             * Reference to the HTML element for progress bar value label.
             *
             * @private
             * @type {HTMLElement}
             */
            recordProgressLabelVal = null,

            /**
             * Reference to the HTML element for progress bar maximum value
             * label.
             *
             * @private
             * @type {HTMLElement}
             */
            recordProgressLabelMax = null,

            /**
             * Specifies whether photo or video mode is active. Value 'true'
             * means that photo mode is active, 'false' that video mode is
             * ready.
             *
             * @private
             * @type {boolean}
             */
            photoMode = true,

            /**
             * Specifies if recording state is active.
             *
             * @private
             * @type {boolean}
             */
            recording = false,

            /**
             * Interface represents a stream of media content fetched from a
             * local data source.
             *
             * @private
             * @type {LocalMediaStream}
             */
            cameraStream = null,

            /**
             * Reference to the main popup HTML element.
             *
             * @private
             * @type {HTMLElement}
             */
            alert = null,

            /**
             * Reference to the container HTML element for main popup message
             * text.
             *
             * @private
             * @type {HTMLElement}
             */
            alertMessage = null,

            /**
             * Reference to the main popup 'OK' button HTML element.
             *
             * @private
             * @type {HTMLElement}
             */
            alertOk = null,

            /**
             * Recording interval ID.
             *
             * @private
             * @type {Number}
             */
            recordingInterval = null,

            /**
             * The maximum length of the recorded video in seconds.
             *
             * @private
             * @type {number}
             */
            maxRecordingTimeSeconds = Math.floor(
                camera.MAX_RECORDING_TIME / 1000
            ),

            /**
             * Current recording time.
             *
             * @private
             * @type {number}
             */
            recordingTime = 0,

            /**
             * Specifies if application exit is in progress.
             *
             * @private
             * @type {boolean}
             */
            exitInProgress = false,

            /**
             * Specifies if tap on the preview layer is allowed.
             *
             * @private
             * @type {boolean}
             */
            previewTapAllowed = false,

            /**
             * Specifies if application should be closed after popup close.
             *
             * @private
             * @type {boolean}
             */
            popupExitAppOnClose = false,

            /**
             * Specifies if starting camera preview is in progress.
             *
             * @private
             * @type {boolean}
             */
            previewInitInProgress = false,

            /**
             * Specifies how many attempts of obtaining access to camera stream
             * was made.
             *
             * @private
             * @type {number}
             */
            previewInitAttempts = 0,

            /**
             * Specifies if page was changed.
             *
             * @private
             * @type {boolean}
             */
            changePage = false;

        /**
         * Opens page.
         * Handles views.initPage.main.open event.
         *
         * @private
         */
        function open() {
            if (camera.isReady() &&
                    cameraPreview.readyState !== READY_STATE_HAVE_NOTHING) {
                // page will be changed on 'playing' event
                // to prevent black screen blink
                changePage = true;
                cameraPreview.play();
            } else {
                tau.changePage('#main');
            }
        }

        /**
         * Pauses camera preview.
         *
         * @private
         */
        function pausePreview() {
            cameraPreview.pause();
        }

        /**
         * Toggles between recording/no recording state.
         *
         * @private
         * @param {boolean} forceValue
         */
        function toggleRecording(forceValue) {
            if (forceValue !== undefined) {
                recording = !!forceValue;
            } else {
                recording = !recording;
            }
        }

        /**
         * Requests screen awake state.
         *
         * @private
         */
        function requestAwakeState() {
            try {
                power.awake();
            } catch (error) {
                console.error('Unable to set awake state', error);
            }
        }

        /**
         * Handles page show event.
         *
         * @private
         */
        function onPageShow() {
            requestAwakeState();
            previewTapAllowed = true;
            toggleRecording(false);
            if (camera.isReady()) {
                cameraPreview.play();
            }
        }

        /**
         * Toggles app mode photo/cam.
         *
         * @private
         */
        function toggleMode() {
            photoMode = !photoMode;
        }

        /**
         * Shows navigation buttons.
         *
         * @private
         */
        function showNavigationBtns() {
            modeBtn.classList.remove('hidden');
            settingsBtn.classList.remove('hidden');
        }

        /**
         * Hides navigation buttons.
         *
         * @private
         */
        function hideNavigationBtns() {
            modeBtn.classList.add('hidden');
            settingsBtn.classList.add('hidden');
        }

        /**
         * Shows record progress.
         *
         * @private
         */
        function showRecordProgress() {
            recordProgress.classList.remove('hidden');
        }

        /**
         * Hides record progress.
         *
         * @private
         */
        function hideRecordProgress() {
            recordProgress.classList.add('hidden');
        }

        /**
         * Shows recording view.
         *
         * @private
         */
        function showRecordingView() {
            hideNavigationBtns();
            showRecordProgress();
        }

        /**
         * Hides recording view.
         *
         * @private
         */
        function hideRecordingView() {
            showNavigationBtns();
            hideRecordProgress();
        }

        /**
         * Handles page hide event.
         *
         * @private
         */
        function onPageHide() {
            try {
                power.normal();
            } catch (error) {
                console.error('Unable to set normal mode', error);
            }
            hideRecordingView();
            pausePreview();
        }

        /**
         * Updates mode button text.
         *
         * @private
         */
        function updateModeBtn() {
            if (photoMode) {
                modeBtn.classList.remove('movie');
            } else {
                modeBtn.classList.add('movie');
            }
        }

        /**
         * Handles mode button click event.
         *
         * @private
         */
        function onModeBtnClick() {
            toggleMode();
            updateModeBtn();
        }

        /**
         * Handles settings button click event.
         *
         * @private
         */
        function onSettingsBtnClick() {
            e.fire('show.settings', {photoMode: photoMode});
        }

        /**
         * Takes a photo.
         *
         * @private
         */
        function takePhoto() {
            previewTapAllowed = false;
            hideNavigationBtns();

            // ensure preview is visible to prevent webkit freeze
            cameraPreview.classList.remove('hidden');
            camera.takePicture();
        }

        /**
         * Renders recording progress bar value.
         *
         * @private
         * @param {number} value
         */
        function renderRecordingProgressBarValue(value) {
            recordProgressVal.style.width = value + 'px';
        }

        /**
         * Renders recording progress bar label.
         *
         * @private
         */
        function renderRecordingProgressBarLabel() {
            var time = Math.ceil(recordingTime / 1000);

            if (time > maxRecordingTimeSeconds) {
                time = maxRecordingTimeSeconds;
            }
            recordProgressLabelVal.innerHTML =
                dateHelper.formatTime(time);
            recordProgressLabelMax.innerHTML =
                dateHelper.formatTime(maxRecordingTimeSeconds);
        }

        /**
         * Renders recording progress bar.
         *
         * @private
         */
        function renderRecordingProgressBar() {
            var parentWidth = recordProgress.clientWidth,
                width = recordingTime / camera.MAX_RECORDING_TIME * parentWidth;
            renderRecordingProgressBarValue(width);
            renderRecordingProgressBarLabel();
        }

        /**
         * Resets recording progress.
         *
         * @private
         */
        function resetRecordingProgress() {
            recordingTime = 0;
            renderRecordingProgressBar();
        }

        /**
         * Removes recording interval.
         *
         * @private
         */
        function removeRecordingInterval() {
            clearInterval(recordingInterval);
        }

        /**
         * Prevents default action for specified event.
         *
         * @private
         * @param {event} e
         */
        function preventEventDefault(e) {
            e.preventDefault();
        }

        /**
         * Sets and shows status message.
         *
         * @private
         * @param {string} message
         */
        function showStatusMessage(message) {
            var content = document.querySelector('#message .ui-popup-content'),
                container = null;

            content.innerHTML = message;
            tau.openPopup('#message');
            container = page.getElementsByClassName('ui-popup-background')[0];
            if (container) {
                container.removeEventListener('touchstart',
                    preventEventDefault);
                container.addEventListener('touchstart', preventEventDefault);
            }

        }

        /**
         * Clears and hides status message.
         *
         * @private
         */
        function hideStatusMessage() {
            if (window.location.hash === '#?popup=true') {
                tau.closePopup({transition: 'none'});
            }
        }

        /**
         * Updates recording progress.
         *
         * @private
         */
        function updateRecordingProgress() {
            recordingTime = camera.getRecordingTime();

            renderRecordingProgressBar();
        }

        /**
         * Sets recording interval.
         *
         * @private
         */
        function setRecordingInterval() {
            recordingInterval = setInterval(
                updateRecordingProgress,
                RECORDING_INTERVAL_STEP
            );
        }

        /**
         * Starts video recording.
         *
         * @private
         */
        function startRecording() {
            previewTapAllowed = false;
            // ensure preview is visible to prevent webkit freeze.
            cameraPreview.classList.remove('hidden');
            camera.startRecording();
            resetRecordingProgress();
            showRecordingView();
        }

        /**
         * Stops video recording.
         *
         * @private
         */
        function stopRecording() {
            previewTapAllowed = false;
            camera.stopRecording();
        }

        /**
         * Starts or stops video recording.
         *
         * @private
         */
        function setRecording() {
            if (recording) {
                startRecording();
            } else {
                stopRecording();
            }
        }

        /**
         * Handles camera preview click event.
         *
         * @private
         * @param {event} ev
         */
        function onCameraPreviewClick(ev) {
            ev.preventDefault();
            ev.stopPropagation();

            if (!previewTapAllowed) {
                return;
            }

            if (photoMode) {
                takePhoto();
            } else {
                toggleRecording();
                setRecording();
            }
        }

        /**
         * Shows alert popup.
         *
         * @private
         * @param {string} message
         * @param {boolean} exitAppOnClose
         */
        function showAlert(message, exitAppOnClose) {
            popupExitAppOnClose = exitAppOnClose;
            alertMessage.innerHTML = message;
            tau.openPopup('#main-alert');
        }

        /**
         * Gets media stream.
         *
         * @private
         */
        function getUserMedia(onUserMediaSuccess, onUserMediaError) {
            navigator.webkitGetUserMedia(
                {
                    video: true,
                    audio: true
                },
                onUserMediaSuccess,
                onUserMediaError
            );
        }

        /**
         * Executes when preview stream is obtained.
         *
         * @private
         * @param {LocalMediaStream} stream
         */
        function onPreviewStream(stream) {
            var streamUrl = window.webkitURL.createObjectURL(stream);

            previewInitAttempts = 0;
            cameraStream = stream;
            cameraPreview.src = streamUrl;
            camera.registerStream(cameraStream);
            if (pageHelper.isPageActive(page)) {
                cameraPreview.play();
            }
        }

        /**
         * Executes if preview stream error occurs.
         *
         * @private
         */
        function onPreviewStreamError() {
            previewInitAttempts += 1;
            if (previewInitAttempts < PREVIEW_INIT_MAX_ATTEMPTS) {
                // application tries to obtain camera stream up to 3 times
                // because other application may not release it yet
                window.setTimeout(
                    function retry() {
                        getUserMedia(onPreviewStream, onPreviewStreamError);
                    },
                    PREVIEW_INIT_ATTEMPT_TIMEOUT
                );
            } else {
                previewInitAttempts = 0;
                if (document.visibilityState === 'visible') {
                    showAlert(CANNOT_ACCESS_CAMERA_MSG, true);
                }
            }
        }

        /**
         * Requests for camera stream.
         *
         * @private
         */
        function requestForCameraStream() {
            return getUserMedia(onPreviewStream, onPreviewStreamError);
        }

        /**
         * Initiates camera preview.
         *
         * @private
         */
        function initCameraPreview() {
            if (previewInitInProgress) {
                return;
            }

            previewInitInProgress = true;
            requestForCameraStream();
        }

        /**
         * Unregisters listeners for click events.
         *
         * @private
         */
        function unbindCameraReadyEvents() {
            modeBtn.removeEventListener('click', onModeBtnClick);
            settingsBtn.removeEventListener('click', onSettingsBtnClick);
            cameraPreview.removeEventListener('click', onCameraPreviewClick);
            recordProgress.removeEventListener('click', onCameraPreviewClick);
        }

        /**
         * Registers listeners for click events.
         *
         * @private
         */
        function bindCameraReadyEvents() {
            modeBtn.addEventListener('click', onModeBtnClick);
            settingsBtn.addEventListener('click', onSettingsBtnClick);
            cameraPreview.addEventListener('click', onCameraPreviewClick);
            recordProgress.addEventListener('click', onCameraPreviewClick);
        }

        /**
         * Returns true if application is launched on emulator,
         * false otherwise.
         *
         * @private
         * @return {boolean}
         */
        function isDeviceEmulated() {
            return !!(navigator.platform &&
                navigator.platform.indexOf('emulated') !== -1);
        }

        /**
         * Handles models.camera.error event.
         *
         * @private
         * @param {Event} ev
         */
        function onCameraError(ev) {
            var error = ev.detail.error,
                message = '';

            previewInitInProgress = false;
            if (error.code === 0) {
                message = CAMERA_NOT_AVAILABLE_MSG;
                if (isDeviceEmulated()) {
                    message += ' ' + CAMERA_ON_DEVICE_ONLY_MSG;
                }
            } else {
                message = CAMERA_NOT_AVAILABLE_MSG;
            }

            showAlert(message, true);
        }

        /**
         * Handles models.camera.release event.
         *
         * @private
         */
        function onCameraRelease() {
            unbindCameraReadyEvents();
            previewTapAllowed = false;
        }

        /**
         * Handles models.camera.ready event.
         *
         * @private
         */
        function onCameraReady() {
            previewInitInProgress = false;
            if (pageHelper.isPageActive(page)) {
                cameraPreview.play();
            }
        }

        /**
         * Shows autofocus frame.
         *
         * @private
         */
        function showAutoFocus() {
            focusContainer.classList.remove('hidden');
            focusFrame.classList.add('autofocus-animation');
        }

        /**
         * Hides autofocus frame.
         *
         * @private
         */
        function hideAutoFocus() {
            var classList = focusFrame.classList;

            focusContainer.classList.add('hidden');
            classList.remove('autofocus-animation');
            classList.remove('autofocus-success');
            classList.remove('autofocus-failure');
        }

        /**
         * Handles models.camera.shutter event.
         *
         * @private
         */
        function onCameraShutter() {
            hideAutoFocus();
            showStatusMessage('Processing...');
        }

        /**
         * Handles models.camera.autofocus.start event.
         *
         * @private
         */
        function onAutoFocusStart() {
            showAutoFocus();
        }

        /**
         * Handles models.camera.autofocus.success event.
         *
         * @private
         */
        function onAutoFocusSuccess() {
            pausePreview();
            focusFrame.classList.add('autofocus-success');
        }

        /**
         * Handles models.camera.autofocus.failure event.
         *
         * @private
         */
        function onAutoFocusFailure() {
            pausePreview();
            focusFrame.classList.add('autofocus-failure');
        }

        /**
         * Handles models.camera.picture.done event.
         *
         * @private
         * @param {event} ev
         */
        function onPictureDone(ev) {
            var path = ev.detail.path;
            hideAutoFocus();
            e.fire('show.preview', {picture: path});
        }

        /**
         * Handles models.camera.picture.canceled event.
         * Executes when picture taking is canceled.
         *
         * @private
         */
        function onPictureCanceled() {
            hideStatusMessage();
            hideAutoFocus();
        }

        /**
         * Handles models.camera.picture.error event.
         *
         * @private
         * @param {object} error
         */
        function onPictureError(error) {
            console.error('picture error', error);
            hideAutoFocus();
            hideStatusMessage();
        }

        /**
         * Handles models.camera.recording.start event.
         *
         * @private
         */
        function onRecordingStart() {
            setRecordingInterval();
            toggleRecording(true);
            previewTapAllowed = true;
        }

        /**
         * Handles models.camera.recording.done event.
         *
         * @private
         * @param {event} ev
         */
        function onRecordingDone(ev) {
            var path = ev.detail.path;

            removeRecordingInterval();
            toggleRecording(false);
            updateRecordingProgress();
            if (!exitInProgress) {
                e.fire('show.preview', {video: path});
            }
        }

        /**
         * Handles models.camera.recording.error event.
         *
         * @private
         * @param {CustomEvent} ev
         */
        function onRecordingError(ev) {
            var error = ev.detail.error;

            if (error === ERROR_FILE_WRITE) {
                showAlert(NO_FREE_SPACE_MSG);
            } else {
                showAlert(DEFAULT_ERROR_MSG);
            }

            hideRecordingView();
            removeRecordingInterval();
            toggleRecording(false);
        }

        /**
         * Handles canplay event on video element.
         *
         * @private
         * @param {event} ev
         */
        function onVideoCanPlay(ev) {
            setTimeout(function adjustElement() {
                adjustHelper.adjustElement(ev.target);
            }, 0);

            cameraPreview.classList.remove('hidden');
            if (changePage) {
                changePage = false;
                tau.changePage('#main');
            }
        }

        /**
         * Handles playing event on video element.
         *
         * @private
         */
        function onVideoPlaying() {
            bindCameraReadyEvents();
            showNavigationBtns();
        }

        /**
         * Handles camera preview error.
         *
         * @private
         * @param {event} ev
         */
        function onCameraPreviewError(ev) {
            var target = ev.target;
            if (document.visibilityState === 'visible') {
                if (target.networkState === VIDEO_NETWORK_NO_SOURCE) {
                    showAlert(CANNOT_ACCESS_CAMERA_MSG, true);
                } else {
                    console.error('Camera preview error');
                    console.error(ev);
                }
            }
        }

        /**
         * Handles click event on alert OK button.
         *
         * @private
         */
        function closeAlert() {
            if (popupExitAppOnClose) {
                app.exit();
            } else {
                tau.closePopup('#main-alert');
            }
        }

        /**
         * Registers view event listeners.
         *
         * @private
         */
        function bindEvents() {
            previewTapAllowed = true;
            page.addEventListener('pageshow', onPageShow);
            page.addEventListener('pagehide', onPageHide);
            cameraPreview.addEventListener('canplay', onVideoCanPlay);
            cameraPreview.addEventListener('playing', onVideoPlaying);
            cameraPreview.addEventListener('error', onCameraPreviewError);
            alertOk.addEventListener('click', closeAlert);
            alert.addEventListener('popuphide', closeAlert);
        }

        /**
         * Function called when application visibility state changes
         * (document.visibilityState changed to 'visible' or 'hidden').
         * Handles views.initPage.visibility.change event.
         *
         * @private
         */
        function visibilityChange() {
            if (document.visibilityState !== 'visible') {
                if (camera.isReady()) {
                    pausePreview();
                    camera.release();
                    cameraStream.stop();
                }
            } else {
                previewTapAllowed = true;
                if (!cameraStream || cameraStream.ended) {
                    initCameraPreview();
                } else if (!camera.isReady()) {
                    camera.registerStream(cameraStream);
                }
            }
        }

        /**
         * Handles views.initPage.application.exit event.
         *
         * @private
         */
        function onApplicationExit() {
            exitInProgress = true;
            if (camera.isReady()) {
                pausePreview();
                camera.release();
                cameraStream.stop();
            }
        }

        /**
         * Handles views.initPage.application.state.background event.
         *
         * @private
         */
        function onApplicationStateBackground() {
            cameraPreview.classList.add('hidden');
        }

        /**
         * Handles views.initPage.application.state.foreground event.
         *
         * @private
         */
        function onApplicationStateForeground() {
            if (camera.isReady()) {
                cameraPreview.classList.remove('hidden');
            }
        }

        /**
         * Initiates module.
         *
         * @memberof views/main
         * @public
         */
        function init() {
            page = document.getElementById('main');
            modeBtn = document.getElementById('mode-btn');
            settingsBtn = document.getElementById('settings-btn');
            cameraPreview = document.getElementById('camera-preview');
            focusContainer = document.getElementById('focus-container');
            focusFrame = document.getElementById('focus-frame');
            recordProgress = document.getElementById('record-progress');
            recordProgressVal = document.getElementById('record-progress-val');
            recordProgressLabelVal =
                document.getElementById('record-progress-label-val');
            recordProgressLabelMax =
                document.getElementById('record-progress-label-max');
            alert = document.getElementById('main-alert');
            alertMessage = document.getElementById('main-alert-message');
            alertOk = document.getElementById('main-alert-ok');

            if (!camera.isSupported()) {
                window.alert('The application is not supported ' +
                    'on your device (no required API). It will be closed.');
                app.exit();
            }

            bindEvents();
            updateModeBtn();
            initCameraPreview();
            requestAwakeState();
        }

        e.listeners({
            'views.initPage.application.exit': onApplicationExit,

            'models.camera.ready': onCameraReady,
            'models.camera.release': onCameraRelease,
            'models.camera.error': onCameraError,

            'models.camera.shutter': onCameraShutter,
            'models.camera.picture.done': onPictureDone,
            'models.camera.picture.error': onPictureError,
            'models.camera.picture.canceled': onPictureCanceled,
            'models.camera.autofocus.start': onAutoFocusStart,
            'models.camera.autofocus.success': onAutoFocusSuccess,
            'models.camera.autofocus.failure': onAutoFocusFailure,

            'models.camera.recording.start': onRecordingStart,
            'models.camera.recording.done': onRecordingDone,
            'models.camera.recording.error': onRecordingError,

            'views.initPage.visibility.change': visibilityChange,
            'views.initPage.application.state.background':
                onApplicationStateBackground,
            'views.initPage.application.state.foreground':
                onApplicationStateForeground,
            'views.initPage.main.open': open
        });

        return {
            init: init
        };
    }

});
