/*
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* global define, document, tau */
/* jslint plusplus: true */

/**
 * Recording format page module.
 *
 * @requires core/event
 * @requires core/template
 * @requires models/camera
 * @namespace views/recordingFormat
 * @memberof views
 */

define({
    name: 'views/recordingFormat',
    requires: [
        'core/event',
        'core/template',
        'models/camera'
    ],
    def: function viewsRecordingFormat(req) {
        'use strict';

        /**
         * Core event module.
         *
         * @private
         * @type {Module}
         */
        var e = req.core.event,

            /**
             * Core template module.
             *
             * @private
             * @type {Module}
             */
            t = req.core.template,

            /**
             * Camera models module.
             *
             * @private
             * @type {Module}
             */
            camera = req.models.camera,

            /**
             * Reference to the application recording format page.
             *
             * @private
             * @type {HTMLElement}
             */
            page = null,

            /**
             * Reference to the HTML list element.
             *
             * @private
             * @type {HTMLElement}
             */
            formatList = null;

        /**
         * Handles views.settings.show.recordingFormat.
         *
         * @private
         */
        function show() {
            tau.changePage('#recording-format');
        }

        /**
         * Renders page UI.
         *
         * @private
         */
        function renderView() {
            var formats = camera.getAvailableRecordingFormats(),
                currentFormat = camera.getRecordingFormat(),
                i = 0,
                len = formats.length,
                content = [],
                format = null,
                checked = false;

            for (i = 0; i < len; i += 1) {
                format = formats[i];
                checked = currentFormat === format;
                content.push(t.get('recordingFormatRow', {
                    format: format,
                    checked: checked
                }));
            }

            formatList.innerHTML = content.join('');
        }

        /**
         * Handles pagebeforeshow event.
         *
         * @private
         */
        function onPageBeforeShow() {
            renderView();
        }

        /**
         * Handles click event on format list.
         *
         * @private
         */
        function onFormatListClick() {
            var input = formatList.querySelectorAll('input:checked')[0];

            camera.setRecordingFormat(input.getAttribute('data-format'));
        }

        /**
         * Registers view event listeners.
         *
         * @private
         */
        function bindEvents() {
            page.addEventListener('pagebeforeshow', onPageBeforeShow);
            formatList.addEventListener('click', onFormatListClick);
        }

        /**
         * Initializes module.
         *
         * @memberof views/recordingFormat
         * @public
         */
        function init() {
            page = document.getElementById('recording-format');
            formatList = document.getElementById('recording-format-list');
            bindEvents();
        }

        e.listeners({
            'views.settings.show.recordingFormat': show
        });

        return {
            init: init
        };
    }

});
